// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_HASHTABLE_H
#define OB_HASHTABLE_H

namespace OB
{

//
// The HashtableNode template
//
template<class K, class T>
struct HashtableNode
{
    K key;
    T value;
    CORBA::ULong fullHash;
    HashtableNode<K, T>* next;

    HashtableNode()
    {
    }

    HashtableNode(K k, T val, HashtableNode<K, T>* n = 0)
        : key(k), value(val), next(n)
    {
        fullHash = OBHash(k);
    }
};

//
// The HashEnumerator template
//
template<class K>
class HashEnumerator
{
   K* keys_;
   CORBA::ULong length_;
   CORBA::ULong *ref_;

   void dec()
   {
       if(ref_ != 0)
       {
           (*ref_)--;
           if(*ref_ == 0)
           {
               delete []keys_;
               delete ref_;
           }
       }
   }

public:
   HashEnumerator() : keys_(0), length_(0), ref_(0)
   {
   }

   HashEnumerator(K* keys, CORBA::ULong len)
       : keys_(keys), length_(len)
   {
       ref_ = new CORBA::ULong(0);
       *ref_ = 1;
   }

   HashEnumerator(const HashEnumerator<K>& e)
   {
       keys_ = e.keys_;
       length_ = e.length_;
       ref_ = e.ref_;
       (*ref_)++;
   }

   ~HashEnumerator()
   {
       dec();
   }

   HashEnumerator<K>& operator=(const HashEnumerator<K>& e)
   {
       if(keys_ != e.keys_)
       {
           dec();
           keys_ = e.keys_;
           length_ = e.length_;
           ref_ = e.ref_;
           (*ref_)++;
       }
       return *this;
   }

   CORBA::ULong length() const
   {
       return length_;
   }

   K operator[](CORBA::ULong i)
   {
       assert(i < length_);
       return keys_[i];
   }
};

//
// The Hashtable template
//
template<class K, class T>
class Hashtable OB_COLON_MUTEX
{
    typedef HashtableNode<K, T> node;

    //
    // Hashtable table data
    //
    node** table_;
    CORBA::ULong length_;
    CORBA::Float load_;
    CORBA::ULong count_;
    CORBA::ULong threshold_;

    void removeAll();
    void resize();

public:

    typedef HashEnumerator<K> Enumerator;

    //
    // Construct a table with len capacity
    //
    Hashtable(CORBA::ULong len, CORBA::Float load = 0.75);

    //
    // The destructor
    //
    ~Hashtable();

    //
    // Clear the hash table
    //
    void clear();

    //
    // Returns true if the table contain the object
    //
    bool contains(const K& key) const;

    //
    // Put this object, with this key into the table
    //
    void put(const K& key, T value);

    //
    // Remove the entry with the provided key
    //
    void remove(const K& key);

    //
    // Get an object from the table given key
    //
    bool get(const K& key, T& value) const;

    //
    // Return all elements in the table
    //
    Enumerator keys() const;

    //
    // Display stats on the hash table
    //
    void displayStats(OB_STD(ostream)&);
};

} // End of namespace OB

#if defined(HAVE_NO_EXPLICIT_TEMPLATES) && !defined(HAVE_PRAGMA_DEFINE)
#   include <OB/HashtableI.h>
#endif

#endif
